# ##########################################################################
# Copyright (C) 2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
# ##########################################################################

import argparse
import collections
import csv
import math
import os
import re
import shlex
import sys
from itertools import chain, repeat
from subprocess import Popen, PIPE


#################################################
######### Benchmark suites definitions ##########
#################################################
common = '--iters 3 --perf 1' #always do 3 iterations in perf mode

"""
SYEVD tests are run, for the given precision and sizes, with vectors and without vectors
"""
def syevd_heevd_suite(*, suite, precision, sizenormal, sizebatch):
    fn = 'syevd' if precision == 's' or precision == 'd' else 'heevd'
    size = sizenormal
    for v in ['V', 'N']:
        if v == 'V': vv = 'yes'
        else: vv = 'no'
        for s in size:
            row = {'name': precision+suite, 'name_test': suite, 'function': fn, 'precision': precision, 'evect': vv, 'n': s}
            yield (row, s, f'-f {fn} -r {precision} --evect {v} -n {s} {common}')

"""
SYEVDX tests are run, for the given precision and sizes, with vectors and without vectors and
computing 20, 60 and 100 percent of the eigenvalues
"""
def syevdx_heevdx_suite(*, suite, precision, sizenormal, sizebatch):
    fn = 'syevdx' if precision == 's' or precision == 'd' else 'heevdx'
    size=sizenormal
    for per in [20, 60, 100]:
        for v in ['V', 'N']:
            if v == 'V': vv = 'yes'
            else: vv = 'no'
            for s in size:
                p = int(s * per / 100)
                if p == 0: p = 1
                row = {'name': precision+suite, 'name_test': suite, 'function': fn, 'precision': precision, 'range': per, 'evect': vv, 'n': s}
                yield (row, s, f'-f {fn} -r {precision} --erange I --il 1 --iu {p} --evect {v} -n {s} {common}')

"""
SYEVJ tests are run, for the given precision and sizes, with vectors and without vectors
"""
def syevj_heevj_suite(*, suite, precision, sizenormal, sizebatch):
    fn = 'syevj' if precision == 's' or precision == 'd' else 'heevj'
    size = sizenormal
    for v in ['V', 'N']:
        if v == 'V': vv = 'yes'
        else: vv = 'no'
        for s in size:
            row = {'name': precision+suite, 'name_test': suite, 'function': fn, 'precision': precision, 'evect': vv, 'n': s}
            yield (row, s, f'-f {fn} -r {precision} --evect {v} -n {s} {common}')

"""
SYEVJBATCH tests are run, for the given precision and sizes, with vectors and without vectors
"""
def syevj_heevjBatch_suite(*, suite, precision, sizenormal, sizebatch):
    fn = 'syevj_strided_batched' if precision == 's' or precision == 'd' else 'heevj_strided_batched'
    size = sizebatch
    for v in ['V', 'N']:
        if v == 'V': vv = 'yes'
        else: vv = 'no'
        for s, bc in size:
            row = {'name': precision+suite, 'name_test': suite, 'function': fn, 'precision': precision, 'evect': vv, 'batch_count': bc, 'n': s}
            yield (row, s, f'-f {fn} -r {precision} --evect {v} --batch_count {bc} -n {s} {common}')

"""
GESVD tests are run, for the given precision and sizes, with vectors and without vectors
"""
def gesvd_suite(*, suite, precision, sizenormal, sizebatch):
    fn = 'gesvd'
    size = sizenormal
    for v in ['V', 'N']:
        if v == 'V': vv = 'yes'
        else: vv = 'no'
        for s in size:
            row = {'name': precision+suite, 'name_test': suite, 'function': fn, 'precision': precision, 'svect': vv, 'n': s}
            yield (row, s, f'-f {fn} -r {precision} --left_svect {v} --right_svect {v} -m {s} {common}')

"""
GESVDJ tests are run, for the given precision and sizes, with vectors and without vectors
"""
def gesvdj_suite(*, suite, precision, sizenormal, sizebatch):
    fn = 'gesvdj'
    size = sizenormal
    for v in ['V', 'N']:
        if v == 'V': vv = 'yes'
        else: vv = 'no'
        for s in size:
            row = {'name': precision+suite, 'name_test': suite, 'function': fn, 'precision': precision, 'svect': vv, 'n': s}
            yield (row, s, f'-f {fn} -r {precision} --left_svect {v} --right_svect {v} -m {s} {common}')

"""
GESVDJBATCH tests are run, for the given precision and sizes, with vectors and without vectors
"""
def gesvdjBatch_suite(*, suite, precision, sizenormal, sizebatch):
    fn = 'gesvdj_strided_batched'
    size = sizebatch
    for v in ['V', 'N']:
        if v == 'V': vv = 'yes'
        else: vv = 'no'
        for s, bc in size:
            row = {'name': precision+suite, 'name_test': suite, 'function': fn, 'precision': precision, 'evect': vv, 'batch_count': bc, 'n': s}
            yield (row, s, f'-f {fn} -r {precision} --left_svect {v} --right_svect {v} --batch_count {bc} -m {s} {common}')

"""
POTRF tests are run with the given precision and sizes
"""
def potrf_suite(*, suite, precision, sizenormal, sizebatch):
    fn = 'potrf'
    size = sizenormal
    for s in size:
        row = {'name': precision+suite, 'name_test': suite, 'function': fn, 'precision': precision, 'n': s}
        yield (row, s, f'-f {fn} -r {precision} -n {s} {common}')

"""
POTRFBATCH tests are run with the given precision and sizes
"""
def potrfBatch_suite(*, suite, precision, sizenormal, sizebatch):
    fn = 'potrf_batched'
    size = sizebatch
    for s, bc in size:
        row = {'name': precision+suite, 'name_test': suite, 'function': fn, 'precision': precision, 'batch_count': bc, 'n': s}
        yield (row, s, f'-f {fn} -r {precision} --batch_count {bc} -n {s} {common}')

"""
GEQRF tests are run, for the given precision and number of rows,
with 160, 576, 1088, 2176, and 4352 columns and also for the square case (#rows = #columns)
"""
def geqrf_suite(*, suite, precision, sizenormal, sizebatch):
    fn = 'geqrf'
    size=sizenormal
    for nc in [0, 160, 576, 1088, 2176, 4352]:
        if nc == 0: nn = 'sq'
        else: nn = nc
        for s in size:
            if nc == 0: n = s
            else: n = nc
            row = {'name': precision+suite, 'name_test': suite, 'function': fn, 'precision': precision, 'cols': nn, 'n': s}
            yield (row, s, f'-f {fn} -r {precision} -n {n} -m {s} {common}')

suites = {
  'syevd': syevd_heevd_suite,
  'syevdx': syevdx_heevdx_suite,
  'syevj': syevj_heevj_suite,
  'syevjBatch': syevj_heevjBatch_suite,
  'gesvd': gesvd_suite,
  'gesvdj': gesvdj_suite,
  'gesvdjBatch': gesvdjBatch_suite,
  'potrf': potrf_suite,
  'potrfBatch': potrfBatch_suite,
  'geqrf': geqrf_suite}


#################################################
############## Helper functions #################
#################################################
"""
SETUP_VPRINT defines the function vprint as the normal print function when
verbose output is enabled, or alternatively as a function that does nothing.
"""
def setup_vprint(args):
    global vprint
    vprint = print if args.verbose else lambda *a, **k: None

"""
CALL_ROCSOLVER_BENCH executes system call to the benchmark
client executable with the given list of arguments
"""
def call_rocsolver_bench(bench_executable, *args):
    cmd = [bench_executable]
    for arg in args:
        if isinstance(arg, str):
            cmd.extend(shlex.split(arg, False, False))
        elif isinstance(arg, collections.Sequence):
            cmd.extend(arg)
        else:
            cmd.push(str(arg))
    process = Popen(cmd, stdout=PIPE, stderr=PIPE)
    vprint('executing {}'.format(' '.join(cmd)))
    stdout, stderr = process.communicate()
    return (str(stdout, encoding='utf-8', errors='surrogateescape'),
            str(stderr, encoding='utf-8', errors='surrogateescape'),
            process.returncode)

"""
EXECUTE_BENCHMARKS collects the arguments for the benchmark client, calls
the client, gets the resulting time, and put everything in file or screen
"""
def execute_benchmarks(output_file, suite, precision, case, bench_executable):
    init = False
    benchmark_generator = suites[suite];
    sizenormal = list(chain(range(2, 64, 8), range(64, 256, 32), range(256, 1024, 64)))
    sizebatch = list(chain(zip(range(2, 64, 4), repeat(5000)), zip(range(72, 164, 8), repeat(2500))))
    if case == 'medium' or case == 'large':
        sizenormal += list(chain(range(1024, 2048, 64), range(2048, 4096, 128)))
        sizebatch += list(chain(zip(range(168, 260, 8), repeat(2500)), zip(range(272, 520, 16), repeat(1000))))
    if case == 'large':
        sizenormal += list(chain(range(4096, 8192, 256), range(8192, 12300, 512)))
        sizebatch += list(chain(zip(range(544, 1050, 32), repeat(500)), zip(range(1088, 2050, 64), repeat(50))))

    for row, n, bench_args in benchmark_generator(suite=suite, precision=precision, sizenormal=sizenormal, sizebatch=sizebatch):
        out, err, exitcode = call_rocsolver_bench(bench_executable, bench_args)
        if exitcode != 0:
            sys.exit("rocsolver-bench call failure: {}".format(err))
        time = float(out)
        row['gpu_time_us'] = time
        row['log_n'] = math.log10(n)
        row['log_gpu_time_us'] = math.log10(time)
        if not init:
            results = csv.DictWriter(output_file, fieldnames=row.keys(), extrasaction='raise', dialect='excel')
            results.writeheader()
            init = True
        results.writerow(row)


#################################################
######### Main functions ########################
#################################################
if __name__ == '__main__':
    parser = argparse.ArgumentParser(prog='rocsolver-perfoptim-suite',
            description='Executes a selected suite of benchmarks and collates the results.')
    parser.add_argument('-v','--verbose',
            action='store_true',
            help='display more information about operations being performed')
    parser.add_argument('--exe',
            default='../../build/release/clients/staging/rocsolver-bench',
            help='the benchmark executable to run')
    parser.add_argument('-o',
            dest='output_path',
            default=None,
            help='the output file name for the benchmark results')
    parser.add_argument('suite',
            choices=suites.keys(),
            help='the set of benchmarks to run')
    parser.add_argument('precision',
            choices=['s', 'd', 'c' , 'z'],
            help='the precision to use for the benchmarks')
    parser.add_argument('case',
            choices=['small', 'medium', 'large'],
            help='the size case to use for the benchmarks')
    args = parser.parse_args()
    setup_vprint(args)

    if args.output_path is not None:
        with open(args.output_path, 'w', buffering=1, encoding='utf-8') as output_file:
            execute_benchmarks(output_file, args.suite, args.precision, args.case, args.exe)
    else:
        execute_benchmarks(sys.stdout, args.suite, args.precision, args.case, args.exe)





# ##########################################################################
# Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
# ##########################################################################

import argparse
import pandas as pd

def append_units(label):
     if label.endswith('_time'):
         return label + '_us'
     else:
         return label

def operation_count(n, function):
    """Returns the number of operations required for a given function with an n-by-n matrix."""
    if function.startswith('getrf'):
        return (2 * n ** 3 / 3) - (n ** 2 / 2) - (n / 6)
    elif function.startswith('geqrf'):
        return (4 * n ** 3 / 3) + (2 * n ** 2) + (4 * n)
    elif function.startswith('getri'):
        return (4 * n ** 3 / 3) - (n ** 2) + (5 * n / 3)
    else: 
        raise ValueError(f'Could not calculate operation count for {function}.')

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
            description='Processes the rocsolver-benchmark-suite results.')
    parser.add_argument('input_path',
            help='the input file name for the raw benchmark results')
    parser.add_argument('output_path',
            help='the output file name for the processed benchmark results')
    args = parser.parse_args()

    # load the data and filter on the function name
    df = pd.read_csv(args.input_path, encoding='utf-8')

    # rename columns for nicer labels
    df.rename(append_units, axis='columns', inplace=True)

    # prepend name column
    df.insert(0, 'name', df['precision'].str.cat(df['function']))

    # the op_count derived column calculation can't handle a mix of functions
    if not df['function'].empty and (df['function'] != df['function'][0]).all():
        raise ValueError('This script can only handle one function at a time.')
    function = df['function'][0]

    # append derived columns
    df['mean_gpu_time_us_per_matrix'] = df['gpu_time_us'].astype(float) / df.get('batch_c', 1)
    df['op_count'] = operation_count(df['n'], function).round().astype(int) * df.get('batch_c', 1)
    df['performance_gflops'] = df['op_count'] / df['gpu_time_us'] / 1000

    # save results
    df.to_csv(args.output_path, encoding='utf-8', index=False)



# ##########################################################################
# Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
# ##########################################################################

import collections
import os
import re
import shlex
import unittest
from subprocess import Popen, PIPE

def call_rocsolver_bench(*args):
    cmd = ['./rocsolver-bench']
    for arg in args:
        if isinstance(arg, str):
            cmd.extend(shlex.split(arg, False, False))
        elif isinstance(arg, collections.Sequence):
            cmd.extend(arg)
        else:
            cmd.push(str(arg))
    process = Popen(cmd, stdout=PIPE, stderr=PIPE)
    stdout, stderr = process.communicate()
    return (str(stdout, encoding='utf-8', errors='surrogateescape'),
            str(stderr, encoding='utf-8', errors='surrogateescape'),
            process.returncode)

class TestRocsolverBench(unittest.TestCase):
    def parse_arguments(self, bench_output):
        m = re.search(r"\n=+\nArguments:\s*\n=+\n(?P<arg_names>.*)\n(?P<arg_values>.*)\n",
                bench_output, re.MULTILINE)
        self.assertTrue(m)
        arg_names = m.group('arg_names').split()
        arg_values = m.group('arg_values').split()
        self.assertEqual(len(arg_names), len(arg_values))
        return dict(zip(arg_names, arg_values))

    def parse_results(self, bench_output):
        m = re.search(r"\n=+\nResults:\s*\n=+\n(?P<arg_names>.*)\n(?P<arg_values>.*)\n",
                bench_output, re.MULTILINE)
        self.assertTrue(m)
        arg_names = m.group('arg_names').split()
        arg_values = m.group('arg_values').split()
        self.assertEqual(len(arg_names), len(arg_values))
        return dict(zip(arg_names, arg_values))

    def test_help(self):
        out, err, exitcode = call_rocsolver_bench('--help')
        self.assertEqual(err, '')
        self.assertEqual(exitcode, 0)
        self.assertNotEqual(out, '')

    def test_validate_precision(self):
        for precision in 'sdcz':
            with self.subTest(precision=precision):
                out, err, exitcode = call_rocsolver_bench(f'-f gels -r {precision} -n 10 -m 15')
                self.assertEqual(err, '')
                self.assertEqual(exitcode, 0)

        out, err, exitcode = call_rocsolver_bench('-f gels -r 0 -n 10 -m 15')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_validate_operation(self):
        for trans in 'NTC':
            with self.subTest(trans=trans):
                out, err, exitcode = call_rocsolver_bench(f'-f gels --trans {trans} -n 10 -m 15')
                self.assertEqual(err, '')
                self.assertEqual(exitcode, 0)

        out, err, exitcode = call_rocsolver_bench('-f gels --trans 0 -n 10 -m 15')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_validate_side(self):
        for side in 'LRB':
            with self.subTest(side=side):
                out, err, exitcode = call_rocsolver_bench(f'-f larf --side {side} -n 10 -m 15')
                self.assertEqual(err, '')
                self.assertEqual(exitcode, 0)

        out, err, exitcode = call_rocsolver_bench('-f larf --side 0 -n 10 -m 15')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_validate_fill(self):
        for fill in 'ULF':
            with self.subTest(fill=fill):
                out, err, exitcode = call_rocsolver_bench(f'-f bdsqr --uplo {fill} -n 15 --nu 10 --nv 10')
                self.assertEqual(err, '')
                self.assertEqual(exitcode, 0)

        out, err, exitcode = call_rocsolver_bench('-f bdsqr --uplo 0 -n 15 --nu 10 --nv 10')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_validate_diagonal(self):
        for diag in 'NU':
            with self.subTest(diag=diag):
                out, err, exitcode = call_rocsolver_bench(f'-f trtri --diag {diag} -n 10')
                self.assertEqual(err, '')
                self.assertEqual(exitcode, 0)

        out, err, exitcode = call_rocsolver_bench('-f trtri --diag 0 -n 10')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_validate_direct(self):
        for direct in 'FB':
            with self.subTest(direct=direct):
                out, err, exitcode = call_rocsolver_bench(f'-f larfb --direct {direct} --side L --storev C -n 10 -m 15 -k 1')
                self.assertEqual(err, '')
                self.assertEqual(exitcode, 0)

        out, err, exitcode = call_rocsolver_bench('-f larfb --direct 0 --side L --storev C -k 1 -n 10 -m 15 -k 1')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_validate_storev(self):
        for storev in 'CR':
            with self.subTest(storev=storev):
                out, err, exitcode = call_rocsolver_bench(f'-f larft --storev {storev} -n 10 -k 1')
                self.assertEqual(err, '')
                self.assertEqual(exitcode, 0)

        out, err, exitcode = call_rocsolver_bench('-f larft --storev 0 -n 10 -k 1')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_validate_left_svect(self):
        for svect in 'ASON':
            with self.subTest(svect=svect):
                out, err, exitcode = call_rocsolver_bench(f'-f gesvd --left_svect {svect} -n 10 -m 15')
                self.assertEqual(err, '')
                self.assertEqual(exitcode, 0)

        out, err, exitcode = call_rocsolver_bench('-f gesvd --left_svect 0 -n 10 -m 15')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_validate_right_svect(self):
        for svect in 'ASON':
            with self.subTest(svect=svect):
                out, err, exitcode = call_rocsolver_bench(f'-f gesvd --right_svect {svect} -n 10 -m 15')
                self.assertEqual(err, '')
                self.assertEqual(exitcode, 0)

        out, err, exitcode = call_rocsolver_bench('-f gesvd --right_svect 0 -n 10 -m 15')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_validate_workmode(self):
        for workmode in 'IO':
            with self.subTest(workmode=workmode):
                out, err, exitcode = call_rocsolver_bench(f'-f gesvd --fast_alg {workmode} -n 10 -m 15')
                self.assertEqual(err, '')
                self.assertEqual(exitcode, 0)

        out, err, exitcode = call_rocsolver_bench('-f gesvd --fast_alg 0 -n 10 -m 15')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_validate_evect(self):
        for evect in 'VIN':
            with self.subTest(evect=evect):
                out, err, exitcode = call_rocsolver_bench(f'-f syev --evect {evect} -n 10')
                self.assertEqual(err, '')
                self.assertEqual(exitcode, 0)

        out, err, exitcode = call_rocsolver_bench('-f syev --evect 0 -n 10')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_validate_erange(self):
        for erange in 'AVI':
            with self.subTest(erange=erange):
                out, err, exitcode = call_rocsolver_bench(f'-f stebz --erange {erange} -n 10')
                self.assertEqual(err, '')
                self.assertEqual(exitcode, 0)

        out, err, exitcode = call_rocsolver_bench('-f stebz --erange 0 -n 10')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_validate_eorder(self):
        for eorder in 'BE':
            with self.subTest(eorder=eorder):
                out, err, exitcode = call_rocsolver_bench(f'-f stebz --eorder {eorder} -n 10')
                self.assertEqual(err, '')
                self.assertEqual(exitcode, 0)

        out, err, exitcode = call_rocsolver_bench('-f stebz --eorder 0 -n 10')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_validate_itype(self):
        for itype in '123':
            with self.subTest(itype=itype):
                out, err, exitcode = call_rocsolver_bench(f'-f sygv --itype {itype} -n 10')
                self.assertEqual(err, '')
                self.assertEqual(exitcode, 0)

        out, err, exitcode = call_rocsolver_bench('-f sygv --itype 0 -n 10')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_unused_arg(self):
        out, err, exitcode = call_rocsolver_bench('-f gels --itype 1 -n 10 -m 15')
        self.assertNotEqual(err, '')
        self.assertNotEqual(exitcode, 0)

    def test_perf_returns_number(self):
        out, err, exitcode = call_rocsolver_bench('-f gels -n 10 -m 15 --perf 1')
        self.assertEqual(err, '')
        self.assertEqual(exitcode, 0)
        self.assertGreaterEqual(float(out), 0)

def generate_parameterized_test(command_options, expected_args):
    def test_function_output(self):
        out, err, exitcode = call_rocsolver_bench(command_options)
        self.assertEqual(err, '')
        self.assertEqual(exitcode, 0)

        args = self.parse_arguments(out)
        self.assertEqual(args, expected_args)

        results = self.parse_results(out)
        self.assertGreaterEqual(float(results['cpu_time_us']), 0)
        self.assertGreaterEqual(float(results['gpu_time_us']), 0)
    return test_function_output

parameters = [
    (
        'laswp',
        '-f laswp -n 10 --k1 15 --k2 20',
        {
            'n': '10',
            'lda': '20',
            'k1': '15',
            'k2': '20',
            'inc': '1',
        }
    ),
    (
        'larfg',
        '-f larfg -n 10',
        {
            'n': '10',
            'inc': '1',
        }
    ),
    (
        'larf',
        '-f larf -m 10 --side L',
        {
            'side': 'L',
            'm': '10',
            'n': '10',
            'inc': '1',
            'lda': '10',
        }
    ),
    (
        'larft',
        '-f larft -n 10 --storev C -k 5',
        {
            'direct': 'F',
            'storev': 'C',
            'n': '10',
            'k': '5',
            'ldv': '10',
            'ldt': '5',
        }
    ),
    (
        'larfb',
        '-f larfb -n 10 -m 15 --direct F --side L --storev C -k 5',
        {
            'side': 'L',
            'trans': 'N',
            'direct': 'F',
            'storev': 'C',
            'm': '15',
            'n': '10',
            'k': '5',
            'ldv': '15',
            'ldt': '5',
            'lda': '15',
        }
    ),
    (
        'latrd',
        '-f latrd -n 10 -k 5',
        {
            'uplo': 'U',
            'n': '10',
            'k': '5',
            'lda': '10',
            'ldw': '10',
        }
    ),
    (
        'labrd',
        '-f labrd -n 10 -m 15',
        {
            'm': '15',
            'n': '10',
            'nb': '10',
            'lda': '15',
            'ldx': '15',
            'ldy': '10',
        }
    ),
    (
        'bdsqr',
        '-f bdsqr --uplo U -n 15 --nu 10 --nv 10',
        {
            'uplo': 'U',
            'n': '15',
            'nv': '10',
            'nu': '10',
            'nc': '0',
            'ldv': '15',
            'ldu': '10',
            'ldc': '1',
        }
    ),
    (
        'steqr',
        '-f steqr -n 15',
        {
            'evect': 'N',
            'n': '15',
            'ldc': '15',
        }
    ),
    (
        'stedc',
        '-f stedc -n 15',
        {
            'evect': 'N',
            'n': '15',
            'ldc': '15',
        }
    ),
    (
        'stein',
        '-f stein -n 15',
        {
            'n': '15',
            'nev': '5',
            'ldz': '15',
        }
    ),
    (
        'lasyf',
        '-f lasyf -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'nb': '10',
            'lda': '10',
        }
    ),
    (
        'potf2',
        '-f potf2 -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'potf2_batched',
        '-f potf2_batched -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
            'batch_c': '1',
        }
    ),
    (
        'potf2_strided_batched',
        '-f potf2_strided_batched -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'batch_c': '1',
        }
    ),
    (
        'potrf',
        '-f potrf -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'potrf_batched',
        '-f potrf_batched -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
            'batch_c': '1',
        }
    ),
    (
        'potrf_strided_batched',
        '-f potrf_strided_batched -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'batch_c': '1',
        }
    ),
    (
        'potrs',
        '-f potrs -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'nrhs': '10',
            'lda': '10',
            'ldb': '10',
        }
    ),
    (
        'potrs_batched',
        '-f potrs_batched -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'nrhs': '10',
            'lda': '10',
            'ldb': '10',
            'batch_c': '1',
        }
    ),
    (
        'potrs_strided_batched',
        '-f potrs_strided_batched -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'nrhs': '10',
            'lda': '10',
            'ldb': '10',
            'strideA': '100',
            'strideB': '100',
            'batch_c': '1',
        }
    ),
    (
        'posv',
        '-f posv -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'nrhs': '10',
            'lda': '10',
            'ldb': '10',
        }
    ),
    (
        'posv_batched',
        '-f posv_batched -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'nrhs': '10',
            'lda': '10',
            'ldb': '10',
            'batch_c': '1',
        }
    ),
    (
        'posv_strided_batched',
        '-f posv_strided_batched -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'nrhs': '10',
            'lda': '10',
            'ldb': '10',
            'strideA': '100',
            'strideB': '100',
            'batch_c': '1',
        }
    ),
    (
        'potri',
        '-f potri -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'potri_batched',
        '-f potri_batched -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
            'batch_c': '1',
        }
    ),
    (
        'potri_strided_batched',
        '-f potri_strided_batched -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'batch_c': '1',
        }
    ),
    (
        'getf2_npvt',
        '-f getf2_npvt -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'getf2_npvt_batched',
        '-f getf2_npvt_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'batch_c': '1',
        }
    ),
    (
        'getf2_npvt_strided_batched',
        '-f getf2_npvt_strided_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'batch_c': '1',
        }
    ),
    (
        'getrf_npvt',
        '-f getrf_npvt -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'getrf_npvt_batched',
        '-f getrf_npvt_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'batch_c': '1',
        }
    ),
    (
        'getrf_npvt_strided_batched',
        '-f getrf_npvt_strided_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'batch_c': '1',
        }
    ),
    (
        'getrf',
        '-f getrf -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'getrf_batched',
        '-f getrf_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'getrf_strided_batched',
        '-f getrf_strided_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'getf2',
        '-f getf2 -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'getf2_batched',
        '-f getf2_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'getf2_strided_batched',
        '-f getf2_strided_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'geqr2',
        '-f geqr2 -n 10 -m 15',
        {
            'm': '15',
            'n': '10',
            'lda': '15',
        }
    ),
    (
        'geqr2_batched',
        '-f geqr2_batched -n 10 -m 15',
        {
            'm': '15',
            'n': '10',
            'lda': '15',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'geqr2_strided_batched',
        '-f geqr2_strided_batched -n 10 -m 15',
        {
            'm': '15',
            'n': '10',
            'lda': '15',
            'strideA': '150',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'geqrf',
        '-f geqrf -n 10 -m 15',
        {
            'm': '15',
            'n': '10',
            'lda': '15',
        }
    ),
    (
        'geqrf_batched',
        '-f geqrf_batched -n 10 -m 15',
        {
            'm': '15',
            'n': '10',
            'lda': '15',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'geqrf_strided_batched',
        '-f geqrf_strided_batched -n 10 -m 15',
        {
            'm': '15',
            'n': '10',
            'lda': '15',
            'strideA': '150',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'geqrf_ptr_batched',
        '-f geqrf_ptr_batched -n 10 -m 15',
        {
            'm': '15',
            'n': '10',
            'lda': '15',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'gerq2',
        '-f gerq2 -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'gerq2_batched',
        '-f gerq2_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'gerq2_strided_batched',
        '-f gerq2_strided_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'gerqf',
        '-f gerqf -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'gerqf_batched',
        '-f gerqf_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'gerqf_strided_batched',
        '-f gerqf_strided_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'geql2',
        '-f geql2 -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'geql2_batched',
        '-f geql2_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'geql2_strided_batched',
        '-f geql2_strided_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'geqlf',
        '-f geqlf -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'geqlf_batched',
        '-f geqlf_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'geqlf_strided_batched',
        '-f geqlf_strided_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'gelq2',
        '-f gelq2 -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'gelq2_batched',
        '-f gelq2_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'gelq2_strided_batched',
        '-f gelq2_strided_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'gelqf',
        '-f gelqf -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'gelqf_batched',
        '-f gelqf_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'gelqf_strided_batched',
        '-f gelqf_strided_batched -m 10',
        {
            'm': '10',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'getrs',
        '-f getrs -n 10',
        {
            'trans': 'N',
            'n': '10',
            'nrhs': '10',
            'lda': '10',
            'ldb': '10',
        }
    ),
    (
        'getrs_batched',
        '-f getrs_batched -n 10',
        {
            'trans': 'N',
            'n': '10',
            'nrhs': '10',
            'lda': '10',
            'ldb': '10',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'getrs_strided_batched',
        '-f getrs_strided_batched -n 10',
        {
            'trans': 'N',
            'n': '10',
            'nrhs': '10',
            'lda': '10',
            'ldb': '10',
            'strideA': '100',
            'strideP': '10',
            'strideB': '100',
            'batch_c': '1',
        }
    ),
    (
        'gesv',
        '-f gesv -n 10',
        {
            'n': '10',
            'nrhs': '10',
            'lda': '10',
            'ldb': '10',
        }
    ),
    (
        'gesv_batched',
        '-f gesv_batched -n 10',
        {
            'n': '10',
            'nrhs': '10',
            'lda': '10',
            'ldb': '10',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'gesv_strided_batched',
        '-f gesv_strided_batched -n 10',
        {
            'n': '10',
            'nrhs': '10',
            'lda': '10',
            'ldb': '10',
            'strideA': '100',
            'strideP': '10',
            'strideB': '100',
            'batch_c': '1',
        }
    ),
    (
        'gesvd',
        '-f gesvd -n 10 -m 15',
        {
            'left_svect': 'N',
            'right_svect': 'N',
            'm': '15',
            'n': '10',
            'lda': '15',
            'ldu': '15',
            'ldv': '10',
        }
    ),
    (
        'gesvd_batched',
        '-f gesvd_batched -n 10 -m 15',
        {
            'left_svect': 'N',
            'right_svect': 'N',
            'm': '15',
            'n': '10',
            'lda': '15',
            'strideS': '10',
            'ldu': '15',
            'strideU': '225',
            'ldv': '10',
            'strideV': '100',
            'strideE': '9',
            'batch_c': '1',
        }
    ),
    (
        'gesvd_strided_batched',
        '-f gesvd_strided_batched -n 10 -m 15',
        {
            'left_svect': 'N',
            'right_svect': 'N',
            'm': '15',
            'n': '10',
            'lda': '15',
            'strideA': '150',
            'strideS': '10',
            'ldu': '15',
            'strideU': '225',
            'ldv': '10',
            'strideV': '100',
            'strideE': '9',
            'batch_c': '1',
        }
    ),
    (
        'trtri',
        '-f trtri -n 10',
        {
            'uplo': 'U',
            'diag': 'N',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'trtri_batched',
        '-f trtri_batched -n 10',
        {
            'uplo': 'U',
            'diag': 'N',
            'n': '10',
            'lda': '10',
            'batch_c': '1',
        }
    ),
    (
        'trtri_strided_batched',
        '-f trtri_strided_batched -n 10',
        {
            'uplo': 'U',
            'diag': 'N',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'batch_c': '1',
        }
    ),
    (
        'getri',
        '-f getri -n 10',
        {
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'getri_batched',
        '-f getri_batched -n 10',
        {
            'n': '10',
            'lda': '10',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'getri_strided_batched',
        '-f getri_strided_batched -n 10',
        {
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'getri_npvt',
        '-f getri_npvt -n 10',
        {
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'getri_npvt_batched',
        '-f getri_npvt_batched -n 10',
        {
            'n': '10',
            'lda': '10',
            'batch_c': '1',
        }
    ),
    (
        'getri_npvt_strided_batched',
        '-f getri_npvt_strided_batched -n 10',
        {
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'batch_c': '1',
        }
    ),
    (
        'getri_outofplace',
        '-f getri_outofplace -n 10',
        {
            'n': '10',
            'lda': '10',
            'ldc': '10',
        }
    ),
    (
        'getri_outofplace_batched',
        '-f getri_outofplace_batched -n 10',
        {
            'n': '10',
            'lda': '10',
            'strideP': '10',
            'ldc': '10',
            'batch_c': '1',
        }
    ),
    (
        'getri_outofplace_strided_batched',
        '-f getri_outofplace_strided_batched -n 10',
        {
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'strideP': '10',
            'ldc': '10',
            'strideC': '100',
            'batch_c': '1',
        }
    ),
    (
        'getri_npvt_outofplace',
        '-f getri_npvt_outofplace -n 10',
        {
            'n': '10',
            'lda': '10',
            'ldc': '10',
        }
    ),
    (
        'getri_npvt_outofplace_batched',
        '-f getri_npvt_outofplace_batched -n 10',
        {
            'n': '10',
            'lda': '10',
            'ldc': '10',
            'batch_c': '1',
        }
    ),
    (
        'getri_npvt_outofplace_strided_batched',
        '-f getri_npvt_outofplace_strided_batched -n 10',
        {
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'ldc': '10',
            'strideC': '100',
            'batch_c': '1',
        }
    ),
    (
        'gels',
        '-f gels -n 10 -m 15',
        {
            'trans': 'N',
            'm': '15',
            'n': '10',
            'nrhs': '10',
            'lda': '15',
            'ldb': '15',
        }
    ),
    (
        'gels_batched',
        '-f gels_batched -n 10 -m 15',
        {
            'trans': 'N',
            'm': '15',
            'n': '10',
            'nrhs': '10',
            'lda': '15',
            'ldb': '15',
            'batch_c': '1',
        }
    ),
    (
        'gels_strided_batched',
        '-f gels_strided_batched -n 10 -m 15',
        {
            'trans': 'N',
            'm': '15',
            'n': '10',
            'nrhs': '10',
            'lda': '15',
            'ldb': '15',
            'strideA': '150',
            'strideB': '150',
            'batch_c': '1',
        }
    ),
    (
        'gebd2',
        '-f gebd2 -n 10 -m 15',
        {
            'm': '15',
            'n': '10',
            'lda': '15',
        }
    ),
    (
        'gebd2_batched',
        '-f gebd2_batched -n 10 -m 15',
        {
            'm': '15',
            'n': '10',
            'lda': '15',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'gebd2_strided_batched',
        '-f gebd2_strided_batched -n 10 -m 15',
        {
            'm': '15',
            'n': '10',
            'lda': '15',
            'strideA': '150',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'gebrd',
        '-f gebrd -n 10 -m 15',
        {
            'm': '15',
            'n': '10',
            'lda': '15',
        }
    ),
    (
        'gebrd_batched',
        '-f gebrd_batched -n 10 -m 15',
        {
            'm': '15',
            'n': '10',
            'lda': '15',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'gebrd_strided_batched',
        '-f gebrd_strided_batched -n 10 -m 15',
        {
            'm': '15',
            'n': '10',
            'lda': '15',
            'strideA': '150',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'sytf2',
        '-f sytf2 -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'sytf2_batched',
        '-f sytf2_batched -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'sytf2_strided_batched',
        '-f sytf2_strided_batched -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'sytrf',
        '-f sytrf -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
        }
    ),
    (
        'sytrf_batched',
        '-f sytrf_batched -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
            'strideP': '10',
            'batch_c': '1',
        }
    ),
    (
        'sytrf_strided_batched',
        '-f sytrf_strided_batched -n 10',
        {
            'uplo': 'U',
            'n': '10',
            'lda': '10',
            'strideA': '100',
            'strideP': '10',
            'batch_c': '1',
        }
    )
]

if __name__ == '__main__':
    for name, command_options, expected_args in parameters:
        test = generate_parameterized_test(command_options, expected_args)
        setattr(TestRocsolverBench, f'test_{name}', test)
    unittest.main()



#!/usr/bin/python3
"""Copyright 2020-2021 Advanced Micro Devices, Inc.
Manage build and installation"""

import re
import sys
import os
import platform
import subprocess
import argparse
import pathlib
from fnmatch import fnmatchcase

args = {}
param = {}
OS_info = {}

def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(description="""Checks build arguments""")

    # common
    parser.add_argument('-g', '--debug', required=False, default = False,  action='store_true',
                        help='Generate Debug build (optional, default: False)')
    parser.add_argument(      '--build_dir', type=str, required=False, default = "build",
                        help='Build directory path (optional, default: build)')
    parser.add_argument(      '--skip_ld_conf_entry', required=False, default=False)
    parser.add_argument(      '--static', required=False, default = False, dest='static_lib', action='store_true',
                        help='Generate static library build (optional, default: False)')
    parser.add_argument('-n', required=False, default=True, dest='optimal', action='store_false',
                        help='Include specialized kernels for small matrix sizes for better performance. (optional, default: True')
    parser.add_argument('-c', '--clients', required=False, default = False, dest='build_clients', action='store_true',
                        help='Generate all client builds (optional, default: False)')
    parser.add_argument('-i', '--install', required=False, default = False, dest='install', action='store_true',
                        help='Install after build (optional, default: False)')
    parser.add_argument(      '--cmake-darg', required=False, dest='cmake_dargs', action='append', default=[],
                        help='List of additional cmake defines for builds (optional, e.g. CMAKE)')
    parser.add_argument('-v', '--verbose', required=False, default = False, action='store_true',
                        help='Verbose build (optional, default: False)')
    # rocsolver
    parser.add_argument(     '--clients-only', dest='clients_only', required=False, default = False, action='store_true',
                        help='Build only clients with a pre-built library')
    parser.add_argument(     '--rocblas_dir', dest='rocblas_dir', type=str, required=False, default = "",
                        help='Specify path to an existing rocBLAS install directory (optional, default: /opt/rocm/rocblas)')
    parser.add_argument(     '--rocsolver_dir', dest='rocsolver_dir', type=str, required=False, default = "",
                        help='Specify path to an existing rocSOLVER install directory (optional, default: /opt/rocm/rocsolver)')
    parser.add_argument('-a', '--architecture', dest='gpu_architecture', type=str, required=False, default=None,
                        help='Set GPU architectures to build for (optional)')
    parser.add_argument(      '--ci_labels', nargs='?', type=str, required=False, default="",
                        help='Semicolon (";") separated list of GitHub Labels with build options (not implemented, silently ignored)')
    parser.add_argument(      '--ci_gfx', nargs='?', type=str, required=False, default="",
                        help='Semicolon (";") separated list of gfx targets expected on test runs (not implemented, silently ignored)')

    return parser.parse_args()

def os_detect():
    global OS_info
    if os.name == "nt":
        OS_info["ID"] = platform.system()
    else:
        inf_file = "/etc/os-release"
        if os.path.exists(inf_file):
            with open(inf_file) as f:
                for line in f:
                    if "=" in line:
                        k,v = line.strip().split("=")
                        OS_info[k] = v.replace('"','')
    OS_info["NUM_PROC"] = os.cpu_count()
    print(OS_info)

def create_dir(dir_path):
    full_path = ""
    if os.path.isabs(dir_path):
        full_path = dir_path
    else:
        full_path = os.path.join( os.getcwd(), dir_path )
    pathlib.Path(full_path).mkdir(parents=True, exist_ok=True)
    return

def delete_dir(dir_path) :
    if (not os.path.exists(dir_path)):
        return
    if os.name == "nt":
        run_cmd( "RMDIR" , f"/S /Q {dir_path}")
    else:
        run_cmd( "rm" , f"-rf {dir_path}")

def cmake_path(os_path):
    if os.name == "nt":
        return os_path.replace("\\", "/")
    else:
        return os_path

def config_cmd():
    global args
    global OS_info
    cwd_path = os.getcwd()
    cmake_executable = ""
    cmake_options = []
    src_path = cmake_path(cwd_path)
    cmake_platform_opts = []
    cmake_prefix_path = []
    if os.name == "nt":
        # not really rocm path as none exist, HIP_DIR set in toolchain is more important
        rocm_path = os.getenv( 'ROCM_CMAKE_PATH', "C:/github/rocm-cmake-master/share/rocm")
        cmake_executable = "cmake"
        #set CPACK_PACKAGING_INSTALL_PREFIX= defined as blank as it is appended to end of path for archive creation
        cmake_platform_opts.append( "-DCPACK_PACKAGING_INSTALL_PREFIX=" )
        cmake_platform_opts.append( "-DCMAKE_INSTALL_PREFIX=C:/hipSDK" )
        lapack_dir = os.getenv("LAPACK_DIR")
        if lapack_dir is not None:
            cmake_prefix_path.append(lapack_dir)
        cblas_dir = os.getenv("cblas_DIR")
        if cblas_dir is not None:
            cmake_prefix_path.append(cblas_dir)

        if lapack_dir is not None:
            cmake_options.append(f"-DROCSOLVER_LAPACK_PATH={lapack_dir}")
        elif cblas_dir is not None:
            cmake_options.append(f"-DROCSOLVER_LAPACK_PATH={cblas_dir}")

        cmake_platform_opts.append('-DCMAKE_CXX_COMPILER=clang++.exe')
        cmake_platform_opts.append('-DCMAKE_C_COMPILER=clang.exe')
        if os.getenv("NO_VCPKG") is None:
            vcpkg_path =  pathlib.Path(os.getenv("VCPKG_PATH", "C:/github/vcpkg"))
            vcpkg_toolchain = vcpkg_path / 'scripts/buildsystems/vcpkg.cmake'
            cmake_options.append(f"-DCMAKE_TOOLCHAIN_FILE={vcpkg_toolchain.as_posix()} -DVCPKG_TARGET_TRIPLET=x64-windows")
        cmake_options.append("-DCMAKE_STATIC_LIBRARY_SUFFIX=.a")
        cmake_options.append("-DCMAKE_STATIC_LIBRARY_PREFIX=static_")
        cmake_options.append("-DCMAKE_SHARED_LIBRARY_SUFFIX=.dll")
        cmake_options.append("-DCMAKE_SHARED_LIBRARY_PREFIX=")
        cmake_options.append("-G Ninja")
    else:
        rocm_path = os.getenv( 'ROCM_PATH', "/opt/rocm")
        cmake_executable = "cmake"
        cmake_platform_opts.append( f"-DROCM_DIR:PATH={rocm_path} -DCPACK_PACKAGING_INSTALL_PREFIX={rocm_path}" )
        cmake_platform_opts.append("-DCMAKE_INSTALL_PREFIX=rocsolver-install")

    if not args.optimal:
        cmake_options.append('-DOPTIMAL=OFF')

    print( f"Build source path: {src_path}")

    cmake_options.extend( cmake_platform_opts )

    cmake_prefix_path.append(rocm_path)
    cmake_base_options = "-DROCM_PATH={0} -DCMAKE_PREFIX_PATH={1}".format(rocm_path, ';'.join(cmake_prefix_path))
    cmake_options.append( cmake_base_options )

    # packaging options
    cmake_pack_options = f"-DCPACK_SET_DESTDIR=OFF"
    cmake_options.append( cmake_pack_options )

    if os.getenv('CMAKE_CXX_COMPILER_LAUNCHER'):
        cmake_options.append( f"-DCMAKE_CXX_COMPILER_LAUNCHER={os.getenv('CMAKE_CXX_COMPILER_LAUNCHER')}" )

    print( cmake_options )

    # build type
    cmake_config = ""
    build_dir = os.path.abspath(args.build_dir)
    if not args.debug:
        build_path = os.path.join(build_dir, "release")
        cmake_config="Release"
    else:
        build_path = os.path.join(build_dir, "debug")
        cmake_config="Debug"

    cmake_options.append( f"-DCMAKE_BUILD_TYPE={cmake_config}" )

    # clean
    delete_dir( build_path )

    create_dir( os.path.join(build_path, "clients") )
    os.chdir( build_path )

    if args.static_lib:
        cmake_options.append("-DBUILD_SHARED_LIBS=OFF" )

    if args.skip_ld_conf_entry:
        cmake_options.append("-DROCM_DISABLE_LDCONFIG=ON" )

    if args.build_clients:
        cmake_build_dir = cmake_path(build_dir)
        cmake_options.append("-DBUILD_CLIENTS_TESTS=ON -DBUILD_CLIENTS_BENCHMARKS=ON -DBUILD_CLIENTS_SAMPLES=ON" )
        cmake_options.append("-DBUILD_TESTING=ON -DBUILD_CLIENTS_EXTRA_TESTS=ON" )

    if args.clients_only:
        cmake_options.append("-DBUILD_LIBRARY=OFF")

    if args.rocsolver_dir:
        cmake_options.append("-Drocsolver_DIR=" + cmake_path(args.rocsolver_dir))
        os.environ['rocsolver_DIR'] = cmake_path(args.rocsolver_dir)

    if args.rocblas_dir:
        cmake_options.append("-Drocblas_DIR=" + cmake_path(args.rocblas_dir))
        os.environ['rocblas_DIR'] = cmake_path(args.rocblas_dir)

    if args.gpu_architecture is not None:
        cmake_options.append( f"-DAMDGPU_TARGETS={args.gpu_architecture}" )

    if args.cmake_dargs:
        for i in args.cmake_dargs:
          cmake_options.append( f"-D{i}" )

    cmake_options.append( f"{src_path}")
    cmd_opts = " ".join(cmake_options)

    try:
        os.environ['HIP_DIR'] = cmake_path(os.environ['HIP_DIR'])
    except:
        pass

    return cmake_executable, cmd_opts


def make_cmd():
    global args
    global OS_info

    make_options = []

    nproc = OS_info["NUM_PROC"]
    if os.name == "nt":
        make_executable = f"cmake.exe --build . " # ninja
        if args.verbose:
          make_options.append( "--verbose" )
        make_options.append( "--target all" )
        if args.install:
          make_options.append( "--target package --target install" )
    else:
        make_executable = f"make -j{nproc}"
        if args.verbose:
          make_options.append( "VERBOSE=1" )
        if True: # args.install:
         make_options.append( "install" )
    cmd_opts = " ".join(make_options)

    return make_executable, cmd_opts

def run_cmd(exe, opts):
    program = f"{exe} {opts}"
    print(program)
    proc = subprocess.run(program, check=True, stderr=subprocess.STDOUT, shell=True)
    return proc.returncode

def main():
    global args
    os_detect()
    args = parse_args()

    # configure
    exe, opts = config_cmd()
    run_cmd(exe, opts)

    # make
    exe, opts = make_cmd()
    run_cmd(exe, opts)

if __name__ == '__main__':
    main()




# ##########################################################################
# Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
# ##########################################################################

import argparse
import collections
import csv
import math
import os
import re
import shlex
import sys
from itertools import chain, repeat
from subprocess import Popen, PIPE

def setup_vprint(args):
    """
    Defines the function vprint as the normal print function when verbose output
    is enabled, or alternatively as a function that does nothing.
    """
    global vprint
    vprint = print if args.verbose else lambda *a, **k: None

def call_rocsolver_bench(bench_executable, *args):
    cmd = [bench_executable]
    for arg in args:
        if isinstance(arg, str):
            cmd.extend(shlex.split(arg, False, False))
        elif isinstance(arg, collections.Sequence):
            cmd.extend(arg)
        else:
            cmd.push(str(arg))
    process = Popen(cmd, stdout=PIPE, stderr=PIPE)
    vprint('executing {}'.format(' '.join(cmd)))
    stdout, stderr = process.communicate()
    return (str(stdout, encoding='utf-8', errors='surrogateescape'),
            str(stderr, encoding='utf-8', errors='surrogateescape'),
            process.returncode)

class ParseError(Exception):
    pass

def parse_arguments(bench_output):
    m = re.search(r"\n=+\nArguments:\s*\n=+\n(?P<arg_names>.*)\n(?P<arg_values>.*)\n",
            bench_output, re.MULTILINE)
    if not m:
        raise ParseError("Failed to parse arguments")
    arg_names = m.group('arg_names').split()
    arg_values = m.group('arg_values').split()
    if len(arg_names) != len(arg_values):
        raise ParseError("Mismatched argument labels and values")
    return dict(zip(arg_names, arg_values))

def parse_results(bench_output):
    m = re.search(r"\n=+\nResults:\s*\n=+\n(?P<arg_names>.*)\n(?P<arg_values>.*)\n",
            bench_output, re.MULTILINE)
    if not m:
        raise ParseError("Failed to parse results")
    arg_names = m.group('arg_names').split()
    arg_values = m.group('arg_values').split()
    if len(arg_names) != len(arg_values):
        raise ParseError("Mismatched result labels and values")
    return dict(zip(arg_names, arg_values))

def getrf_suite(*, precision):
    for fn in ['getrf', 'getrf_npvt']:
        for m in chain(range(2, 64, 8),
                       range(64, 256, 32),
                       range(256, 2048, 64),
                       range(2048, 4096, 128),
                       range(4096, 8193, 256)):
            yield (fn, f'-f {fn} -r {precision} -m {m} --iters 10')

def getrf_strided_batched_suite(*, precision):
    for fn in ['getrf_strided_batched', 'getrf_npvt_strided_batched']:
        for m, bc in chain(zip(range(2, 64, 1), repeat(5000)),
                           zip(range(64, 256, 8), repeat(2500)),
                           zip(range(256, 384, 16), repeat(1000)),
                           zip(range(384, 512, 32), repeat(750)),
                           zip(range(512, 640, 32), repeat(500)),
                           zip(range(640, 1025, 64), repeat(50))):
            yield (fn, f'-f {fn} -r {precision} -m {m} --iters 10 --batch_count {bc}')

        yield (fn, f'-f {fn} -r {precision} -m 20 --iters 10 --batch_count 4096')
        yield (fn, f'-f {fn} -r {precision} -m 20 --iters 10 --batch_count 32768')

        yield (fn, f'-f {fn} -r {precision} -m 50 --iters 10 --batch_count 4096')
        yield (fn, f'-f {fn} -r {precision} -m 50 --iters 10 --batch_count 32768')

        yield (fn, f'-f {fn} -r {precision} -m 64 --iters 10 --batch_count 1024')
        yield (fn, f'-f {fn} -r {precision} -m 64 --iters 10 --batch_count 2048')
        yield (fn, f'-f {fn} -r {precision} -m 64 --iters 10 --batch_count 4096')

        yield (fn, f'-f {fn} -r {precision} -m 80 --iters 10 --batch_count 4096')
        yield (fn, f'-f {fn} -r {precision} -m 80 --iters 10 --batch_count 32768')

        yield (fn, f'-f {fn} -r {precision} -m 161 --iters 10 --batch_count 1024')
        yield (fn, f'-f {fn} -r {precision} -m 161 --iters 10 --batch_count 2048')
        yield (fn, f'-f {fn} -r {precision} -m 161 --iters 10 --batch_count 4096')

def getri_suite(*, precision):
    for fn in ['getri', 'getri_npvt']:
        for n in chain(range(2, 64, 8),
                       range(64, 256, 32),
                       range(256, 1024, 64),
                       range(1024, 2048, 128),
                       range(2048, 4096, 256),
                       range(4096, 8193, 512)):
            yield (fn, f'-f {fn} -r {precision} -n {n} --iters 10')


def getri_strided_batched_suite(*, precision):
    for fn in ['getri_strided_batched', 'getri_npvt_strided_batched']:
        for n, bc in chain(zip(range(2, 64, 1), repeat(5000)),
                           zip(range(64, 256, 8), repeat(2500)),
                           zip(range(256, 384, 16), repeat(1000)),
                           zip(range(384, 512, 32), repeat(750)),
                           zip(range(512, 640, 32), repeat(500)),
                           zip(range(640, 1025, 64), repeat(50))):
            if precision == 'z' and 232 <= n:
                continue # TODO: fix crash in rocsolver-bench at these sizes
            yield (fn, f'-f {fn} -r {precision} -n {n} --iters 10 --batch_count {bc}')

def geqrf_suite(*, precision):
    for fn in ['geqrf']:
        for m in chain(range(2, 64, 8),
                       range(64, 256, 32),
                       range(256, 1024, 64),
                       range(1024, 2048, 128),
                       range(2048, 4096, 256),
                       range(4096, 8193, 512)):
            yield (fn, f'-f {fn} -r {precision} -m {m} --iters 10')

def geqrf_strided_batched_suite(*, precision):
    for fn in ['geqrf_strided_batched']:
        for m, bc in chain(zip(range(2, 64, 1), repeat(5000)),
                           zip(range(64, 256, 8), repeat(2500)),
                           zip(range(256, 384, 16), repeat(1000)),
                           zip(range(384, 512, 32), repeat(750)),
                           zip(range(512, 640, 32), repeat(500)),
                           zip(range(640, 1025, 64), repeat(50))):
            yield (fn, f'-f {fn} -r {precision} -m {m} --iters 10 --batch_count {bc}')

suites = {
  'geqrf': geqrf_suite,
  'geqrf_strided_batched': geqrf_strided_batched_suite,
  'getrf': getrf_suite,
  'getrf_strided_batched': getrf_strided_batched_suite,
  'getri': getri_suite,
  'getri_strided_batched': getri_strided_batched_suite,
}

def execute_benchmarks(output_file, precision, suite, bench_executable):
    init = False
    benchmark_generator = suites[suite];
    for fn, bench_args in benchmark_generator(precision=precision):
        out, err, exitcode = call_rocsolver_bench(bench_executable, bench_args)
        if exitcode != 0:
            sys.exit("rocsolver-bench call failure: {}".format(err))
        args = parse_arguments(out)
        perf = parse_results(out)
        row = { 'function': fn, 'precision': precision, **args, **perf }
        if not init:
            results = csv.DictWriter(output_file, fieldnames=row.keys(), extrasaction='raise', dialect='excel')
            results.writeheader()
            init = True
        results.writerow(row)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(prog='rocsolver-bench-suite',
            description='Executes a selected suite of benchmarks and collates the results.')
    parser.add_argument('-v','--verbose',
            action='store_true',
            help='display more information about operations being performed')
    parser.add_argument('--exe',
            default='rocsolver-bench',
            help='the benchmark executable to run')
    parser.add_argument('-o',
            dest='output_path',
            default=None,
            help='the output file name for the benchmark results')
    parser.add_argument('precision',
            choices=['s', 'd', 'c' , 'z'],
            help='the precision to use for the benchmark')
    parser.add_argument('suite',
            choices=suites.keys(),
            help='the set of benchmarks to run')
    args = parser.parse_args()
    setup_vprint(args)

    if args.output_path is not None:
        with open(args.output_path, 'w', buffering=1, encoding='utf-8') as output_file:
            execute_benchmarks(output_file, args.precision, args.suite, args.exe)
    else:
        execute_benchmarks(sys.stdout, args.precision, args.suite, args.exe)
